@@ -19,15 +19,15 @@
  *	query plan and ExecutorEnd() should always be called at the end of
  *	execution of a plan.
  *
- *	ExecutorRun accepts 'feature' and 'count' arguments that specify whether
+ *	ExecutorRun accepts direction and count arguments that specify whether
  *	the plan is to be executed forwards, backwards, and for how many tuples.
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/executor/execMain.c,v 1.149 2001/10/25 05:49:27 momjian Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/executor/execMain.c,v 1.180.2.4 2006/01/12 21:49:40 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -35,7 +35,8 @@
 
 #include "access/heapam.h"
 #include "catalog/heap.h"
-#include "commands/command.h"
+#include "catalog/namespace.h"
+#include "commands/tablecmds.h"
 #include "commands/trigger.h"
 #include "executor/execdebug.h"
 #include "executor/execdefs.h"
@@ -43,6 +44,7 @@
 #include "optimizer/var.h"
 #include "parser/parsetree.h"
 #include "utils/acl.h"
+#include "utils/lsyscache.h"
 
 
 /* decls for local routines only used within this module */
@@ -60,15 +62,15 @@ static TupleTableSlot *ExecutePlan(EState *estate, Plan *plan,
 			long numberTuples,
 			ScanDirection direction,
 			DestReceiver *destfunc);
-static void ExecRetrieve(TupleTableSlot *slot,
-			 DestReceiver *destfunc,
-			 EState *estate);
-static void ExecAppend(TupleTableSlot *slot, ItemPointer tupleid,
+static void ExecSelect(TupleTableSlot *slot,
+		   DestReceiver *destfunc,
+		   EState *estate);
+static void ExecInsert(TupleTableSlot *slot, ItemPointer tupleid,
 		   EState *estate);
 static void ExecDelete(TupleTableSlot *slot, ItemPointer tupleid,
 		   EState *estate);
-static void ExecReplace(TupleTableSlot *slot, ItemPointer tupleid,
-			EState *estate);
+static void ExecUpdate(TupleTableSlot *slot, ItemPointer tupleid,
+		   EState *estate);
 static TupleTableSlot *EvalPlanQualNext(EState *estate);
 static void EndEvalPlanQual(EState *estate);
 static void ExecCheckQueryPerms(CmdType operation, Query *parseTree,
@@ -88,7 +90,7 @@ static void ExecCheckRTEPerms(RangeTblEntry *rte, CmdType operation);
  *		query plan
  *
  *		returns a TupleDesc which describes the attributes of the tuples to
- *		be returned by the query.
+ *		be returned by the query.  (Same value is saved in queryDesc)
  *
  * NB: the CurrentMemoryContext when this is called must be the context
  * to be used as the per-query context for the query plan.	ExecutorRun()
@@ -112,22 +114,13 @@ ExecutorStart(QueryDesc *queryDesc, EState *estate)
 	}
 
 	/*
-	 * Make our own private copy of the current queries snapshot data
+	 * Make our own private copy of the current query snapshot data.
+	 *
+	 * This "freezes" our idea of which tuples are good and which are not for
+	 * the life of this query, even if it outlives the current command and
+	 * current snapshot.
 	 */
-	if (QuerySnapshot == NULL)
-		estate->es_snapshot = NULL;
-	else
-	{
-		estate->es_snapshot = (Snapshot) palloc(sizeof(SnapshotData));
-		memcpy(estate->es_snapshot, QuerySnapshot, sizeof(SnapshotData));
-		if (estate->es_snapshot->xcnt > 0)
-		{
-			estate->es_snapshot->xip = (TransactionId *)
-				palloc(estate->es_snapshot->xcnt * sizeof(TransactionId));
-			memcpy(estate->es_snapshot->xip, QuerySnapshot->xip,
-				   estate->es_snapshot->xcnt * sizeof(TransactionId));
-		}
-	}
+	estate->es_snapshot = CopyQuerySnapshot();
 
 	/*
 	 * Initialize the plan
@@ -137,6 +130,8 @@ ExecutorStart(QueryDesc *queryDesc, EState *estate)
 					  queryDesc->plantree,
 					  estate);
 
+	queryDesc->tupDesc = result;
+
 	return result;
 }
 
@@ -149,25 +144,24 @@ ExecutorStart(QueryDesc *queryDesc, EState *estate)
  *
  *		ExecutorStart must have been called already.
  *
- *		the different features supported are:
- *			 EXEC_RUN:	retrieve all tuples in the forward direction
- *			 EXEC_FOR:	retrieve 'count' number of tuples in the forward dir
- *			 EXEC_BACK: retrieve 'count' number of tuples in the backward dir
- *			 EXEC_RETONE: return one tuple but don't 'retrieve' it
- *						   used in postquel function processing
+ *		If direction is NoMovementScanDirection then nothing is done
+ *		except to start up/shut down the destination.  Otherwise,
+ *		we retrieve up to 'count' tuples in the specified direction.
  *
- *		Note: count = 0 is interpreted as "no limit".
+ *		Note: count = 0 is interpreted as no portal limit, e.g. run to
+ *		completion.
  *
  * ----------------------------------------------------------------
  */
 TupleTableSlot *
-ExecutorRun(QueryDesc *queryDesc, EState *estate, int feature, long count)
+ExecutorRun(QueryDesc *queryDesc, EState *estate,
+			ScanDirection direction, long count)
 {
 	CmdType		operation;
 	Plan	   *plan;
-	TupleTableSlot *result;
 	CommandDest dest;
 	DestReceiver *destfunc;
+	TupleTableSlot *result;
 
 	/*
 	 * sanity checks
@@ -181,69 +175,33 @@ ExecutorRun(QueryDesc *queryDesc, EState *estate, int feature, long count)
 	operation = queryDesc->operation;
 	plan = queryDesc->plantree;
 	dest = queryDesc->dest;
-	destfunc = DestToFunction(dest);
+
+	/*
+	 * startup tuple receiver
+	 */
 	estate->es_processed = 0;
 	estate->es_lastoid = InvalidOid;
 
+	destfunc = DestToFunction(dest);
+	(*destfunc->setup) (destfunc, (int) operation,
+						queryDesc->portalName, queryDesc->tupDesc);
+
 	/*
-	 * FIXME: the dest setup function ought to be handed the tuple desc
-	 * for the tuples to be output, but I'm not quite sure how to get that
-	 * info at this point.	For now, passing NULL is OK because no
-	 * existing dest setup function actually uses the pointer.
+	 * run plan
 	 */
-	(*destfunc->setup) (destfunc, (TupleDesc) NULL);
-
-	switch (feature)
-	{
-		case EXEC_RUN:
-			result = ExecutePlan(estate,
-								 plan,
-								 operation,
-								 count,
-								 ForwardScanDirection,
-								 destfunc);
-			break;
-
-		case EXEC_FOR:
-			result = ExecutePlan(estate,
-								 plan,
-								 operation,
-								 count,
-								 ForwardScanDirection,
-								 destfunc);
-			break;
-
-			/*
-			 * retrieve next n "backward" tuples
-			 */
-		case EXEC_BACK:
-			result = ExecutePlan(estate,
-								 plan,
-								 operation,
-								 count,
-								 BackwardScanDirection,
-								 destfunc);
-			break;
-
-			/*
-			 * return one tuple but don't "retrieve" it. (this is used by
-			 * the rule manager..) -cim 9/14/89
-			 */
-		case EXEC_RETONE:
-			result = ExecutePlan(estate,
-								 plan,
-								 operation,
-								 ONE_TUPLE,
-								 ForwardScanDirection,
-								 destfunc);
-			break;
-
-		default:
-			elog(DEBUG, "ExecutorRun: Unknown feature %d", feature);
-			result = NULL;
-			break;
-	}
+	if (direction == NoMovementScanDirection)
+		result = NULL;
+	else
+		result = ExecutePlan(estate,
+							 plan,
+							 operation,
+							 count,
+							 direction,
+							 destfunc);
 
+	/*
+	 * shutdown receiver
+	 */
 	(*destfunc->cleanup) (destfunc);
 
 	return result;
@@ -344,7 +302,7 @@ ExecCheckPlanPerms(Plan *plan, List *rangeTable, CmdType operation)
 
 				/* Recursively check the subquery */
 				rte = rt_fetch(scan->scan.scanrelid, rangeTable);
-				Assert(rte->subquery != NULL);
+				Assert(rte->rtekind == RTE_SUBQUERY);
 				ExecCheckQueryPerms(operation, rte->subquery, scan->subplan);
 				break;
 			}
@@ -391,18 +349,21 @@ ExecCheckRTPerms(List *rangeTable, CmdType operation)
 static void
 ExecCheckRTEPerms(RangeTblEntry *rte, CmdType operation)
 {
-	char	   *relName;
+	Oid			relOid;
 	Oid			userid;
-	int32		aclcheck_result;
+	AclResult	aclcheck_result;
 
 	/*
-	 * If it's a subquery RTE, ignore it --- it will be checked when
-	 * ExecCheckPlanPerms finds the SubqueryScan node for it.
+	 * Only plain-relation RTEs need to be checked here.  Subquery RTEs
+	 * will be checked when ExecCheckPlanPerms finds the SubqueryScan
+	 * node, and function RTEs are checked by init_fcache when the
+	 * function is prepared for execution.	Join and special RTEs need no
+	 * checks.
 	 */
-	if (rte->subquery)
+	if (rte->rtekind != RTE_RELATION)
 		return;
 
-	relName = rte->relname;
+	relOid = rte->relid;
 
 	/*
 	 * userid to check as: current user unless we have a setuid
@@ -415,14 +376,13 @@ ExecCheckRTEPerms(RangeTblEntry *rte, CmdType operation)
 	 */
 	userid = rte->checkAsUser ? rte->checkAsUser : GetUserId();
 
-#define CHECK(MODE)		pg_aclcheck(relName, userid, MODE)
+#define CHECK(MODE)		pg_class_aclcheck(relOid, userid, MODE)
 
 	if (rte->checkForRead)
 	{
 		aclcheck_result = CHECK(ACL_SELECT);
 		if (aclcheck_result != ACLCHECK_OK)
-			elog(ERROR, "%s: %s",
-				 relName, aclcheck_error_strings[aclcheck_result]);
+			aclcheck_error(aclcheck_result, get_rel_name(relOid));
 	}
 
 	if (rte->checkForWrite)
@@ -451,8 +411,7 @@ ExecCheckRTEPerms(RangeTblEntry *rte, CmdType operation)
 				break;
 		}
 		if (aclcheck_result != ACLCHECK_OK)
-			elog(ERROR, "%s: %s",
-				 relName, aclcheck_error_strings[aclcheck_result]);
+			aclcheck_error(aclcheck_result, get_rel_name(relOid));
 	}
 }
 
@@ -491,6 +450,7 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 {
 	List	   *rangeTable;
 	Relation	intoRelationDesc;
+	bool		do_select_into;
 	TupleDesc	tupType;
 
 	/*
@@ -568,6 +528,26 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 		estate->es_result_relation_info = NULL;
 	}
 
+	/*
+	 * Detect whether we're doing SELECT INTO.  If so, set the force_oids
+	 * flag appropriately so that the plan tree will be initialized with
+	 * the correct tuple descriptors.
+	 */
+	do_select_into = false;
+
+	if (operation == CMD_SELECT &&
+		!parseTree->isPortal &&
+		parseTree->into != NULL)
+	{
+		do_select_into = true;
+		/*
+		 * For now, always create OIDs in SELECT INTO; this is for backwards
+		 * compatibility with pre-7.3 behavior.  Eventually we might want
+		 * to allow the user to choose.
+		 */
+		estate->es_force_oids = true;
+	}
+
 	/*
 	 * Have to lock relations selected for update
 	 */
@@ -587,7 +567,7 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 			erm = (execRowMark *) palloc(sizeof(execRowMark));
 			erm->relation = relation;
 			erm->rti = rti;
-			sprintf(erm->resname, "ctid%u", rti);
+			snprintf(erm->resname, 32, "ctid%u", rti);
 			estate->es_rowMark = lappend(estate->es_rowMark, erm);
 		}
 	}
@@ -625,7 +605,7 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 	/*
 	 * Get the tuple descriptor describing the type of tuples to return.
 	 * (this is especially important if we are creating a relation with
-	 * "retrieve into")
+	 * "SELECT INTO")
 	 */
 	tupType = ExecGetTupType(plan);		/* tuple descriptor */
 
@@ -727,23 +707,25 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 	 */
 	intoRelationDesc = (Relation) NULL;
 
-	if (operation == CMD_SELECT)
+	if (do_select_into)
 	{
-		char	   *intoName;
-		Oid			intoRelationId;
-		TupleDesc	tupdesc;
+				char	   *intoName;
+				Oid			namespaceId;
+				AclResult	aclresult;
+				Oid			intoRelationId;
+				TupleDesc	tupdesc;
 
-		if (!parseTree->isPortal)
-		{
-			/*
-			 * a select into table
-			 */
-			if (parseTree->into != NULL)
-			{
 				/*
-				 * create the "into" relation
+				 * find namespace to create in, check permissions
 				 */
-				intoName = parseTree->into;
+				intoName = parseTree->into->relname;
+				namespaceId = RangeVarGetCreationNamespace(parseTree->into);
+
+				aclresult = pg_namespace_aclcheck(namespaceId, GetUserId(),
+												  ACL_CREATE);
+				if (aclresult != ACLCHECK_OK)
+					aclcheck_error(aclresult,
+								   get_namespace_name(namespaceId));
 
 				/*
 				 * have to copy tupType to get rid of constraints
@@ -752,9 +734,10 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 
 				intoRelationId =
 					heap_create_with_catalog(intoName,
+											 namespaceId,
 											 tupdesc,
-											 RELKIND_RELATION, true,
-											 parseTree->isTemp,
+											 RELKIND_RELATION,
+											 false,
 											 allowSystemTableMods);
 
 				FreeTupleDesc(tupdesc);
@@ -771,12 +754,10 @@ InitPlan(CmdType operation, Query *parseTree, Plan *plan, EState *estate)
 				 * with CommandCounterIncrement(), so that the TOAST table
 				 * will be visible for insertion.
 				 */
-				AlterTableCreateToastTable(intoName, true);
+				AlterTableCreateToastTable(intoRelationId, true);
 
 				intoRelationDesc = heap_open(intoRelationId,
 											 AccessExclusiveLock);
-			}
-		}
 	}
 
 	estate->es_into_relation_descriptor = intoRelationDesc;
@@ -822,7 +803,8 @@ initResultRelInfo(ResultRelInfo *resultRelInfo,
 	resultRelInfo->ri_NumIndices = 0;
 	resultRelInfo->ri_IndexRelationDescs = NULL;
 	resultRelInfo->ri_IndexRelationInfo = NULL;
-	resultRelInfo->ri_TrigDesc = resultRelationDesc->trigdesc;
+	/* make a copy so as not to depend on relcache info not changing... */
+	resultRelInfo->ri_TrigDesc = CopyTriggerDesc(resultRelationDesc->trigdesc);
 	resultRelInfo->ri_TrigFunctions = NULL;
 	resultRelInfo->ri_ConstraintExprs = NULL;
 	resultRelInfo->ri_junkFilter = NULL;
@@ -916,10 +898,10 @@ EndPlan(Plan *plan, EState *estate)
  *
  *		processes the query plan to retrieve 'numberTuples' tuples in the
  *		direction specified.
- *		Retrieves all tuples if tupleCount is 0
+ *		Retrieves all tuples if numberTuples is 0
  *
  *		result is either a slot containing the last tuple in the case
- *		of a RETRIEVE or NULL otherwise.
+ *		of a SELECT or NULL otherwise.
  *
  * Note: the ctid attribute is a 'junk' attribute that is removed before the
  * user can see it
@@ -1030,8 +1012,10 @@ lnext:	;
 				foreach(l, estate->es_rowMark)
 				{
 					execRowMark *erm = lfirst(l);
-					Buffer		buffer;
 					HeapTupleData tuple;
+					Buffer		buffer;
+					ItemPointerData update_ctid;
+					TransactionId update_xmax;
 					TupleTableSlot *newSlot;
 					int			test;
 
@@ -1049,22 +1033,32 @@ lnext:	;
 							 erm->resname);
 
 					tuple.t_self = *((ItemPointer) DatumGetPointer(datum));
-					test = heap_mark4update(erm->relation, &tuple, &buffer);
+					test = heap_mark4update(erm->relation, &tuple, &buffer,
+											&update_ctid, &update_xmax,
+											estate->es_snapshot->curcid);
 					ReleaseBuffer(buffer);
 					switch (test)
 					{
 						case HeapTupleSelfUpdated:
+							/* treat it as deleted; do not process */
+							goto lnext;
+
 						case HeapTupleMayBeUpdated:
 							break;
 
 						case HeapTupleUpdated:
 							if (XactIsoLevel == XACT_SERIALIZABLE)
 								elog(ERROR, "Can't serialize access due to concurrent update");
-							if (!(ItemPointerEquals(&(tuple.t_self),
-								  (ItemPointer) DatumGetPointer(datum))))
+							if (!ItemPointerEquals(&update_ctid,
+												   &tuple.t_self))
 							{
-								newSlot = EvalPlanQual(estate, erm->rti, &(tuple.t_self));
-								if (!(TupIsNull(newSlot)))
+								/* updated, so look at updated version */
+								newSlot = EvalPlanQual(estate,
+													   erm->rti,
+													   &update_ctid,
+													   update_xmax,
+													   estate->es_snapshot->curcid);
+								if (!TupIsNull(newSlot))
 								{
 									slot = newSlot;
 									estate->es_useEvalPlan = true;
@@ -1097,26 +1091,25 @@ lnext:	;
 								  InvalidBuffer,		/* this tuple has no
 														 * buffer */
 								  true);		/* tuple should be pfreed */
-		}						/* if (junkfilter... */
+		}
 
 		/*
 		 * now that we have a tuple, do the appropriate thing with it..
 		 * either return it to the user, add it to a relation someplace,
 		 * delete it from a relation, or modify some of its attributes.
 		 */
-
 		switch (operation)
 		{
 			case CMD_SELECT:
-				ExecRetrieve(slot,		/* slot containing tuple */
-							 destfunc,	/* destination's tuple-receiver
+				ExecSelect(slot,	/* slot containing tuple */
+						   destfunc,	/* destination's tuple-receiver
 										 * obj */
-							 estate);	/* */
+						   estate);
 				result = slot;
 				break;
 
 			case CMD_INSERT:
-				ExecAppend(slot, tupleid, estate);
+				ExecInsert(slot, tupleid, estate);
 				result = NULL;
 				break;
 
@@ -1126,12 +1119,12 @@ lnext:	;
 				break;
 
 			case CMD_UPDATE:
-				ExecReplace(slot, tupleid, estate);
+				ExecUpdate(slot, tupleid, estate);
 				result = NULL;
 				break;
 
 			default:
-				elog(DEBUG, "ExecutePlan: unknown operation in queryDesc");
+				elog(LOG, "ExecutePlan: unknown operation in queryDesc");
 				result = NULL;
 				break;
 		}
@@ -1147,25 +1140,25 @@ lnext:	;
 
 	/*
 	 * here, result is either a slot containing a tuple in the case of a
-	 * RETRIEVE or NULL otherwise.
+	 * SELECT or NULL otherwise.
 	 */
 	return result;
 }
 
 /* ----------------------------------------------------------------
- *		ExecRetrieve
+ *		ExecSelect
  *
- *		RETRIEVEs are easy.. we just pass the tuple to the appropriate
+ *		SELECTs are easy.. we just pass the tuple to the appropriate
  *		print function.  The only complexity is when we do a
- *		"retrieve into", in which case we insert the tuple into
+ *		"SELECT INTO", in which case we insert the tuple into
  *		the appropriate relation (note: this is a newly created relation
  *		so we don't need to worry about indices or locks.)
  * ----------------------------------------------------------------
  */
 static void
-ExecRetrieve(TupleTableSlot *slot,
-			 DestReceiver *destfunc,
-			 EState *estate)
+ExecSelect(TupleTableSlot *slot,
+		   DestReceiver *destfunc,
+		   EState *estate)
 {
 	HeapTuple	tuple;
 	TupleDesc	attrtype;
@@ -1181,7 +1174,8 @@ ExecRetrieve(TupleTableSlot *slot,
 	 */
 	if (estate->es_into_relation_descriptor != NULL)
 	{
-		heap_insert(estate->es_into_relation_descriptor, tuple);
+		heap_insert(estate->es_into_relation_descriptor, tuple,
+					estate->es_snapshot->curcid);
 		IncrAppended();
 	}
 
@@ -1194,16 +1188,15 @@ ExecRetrieve(TupleTableSlot *slot,
 }
 
 /* ----------------------------------------------------------------
- *		ExecAppend
+ *		ExecInsert
  *
- *		APPENDs are trickier.. we have to insert the tuple into
+ *		INSERTs are trickier.. we have to insert the tuple into
  *		the base relation and insert appropriate tuples into the
  *		index relations.
  * ----------------------------------------------------------------
  */
-
 static void
-ExecAppend(TupleTableSlot *slot,
+ExecInsert(TupleTableSlot *slot,
 		   ItemPointer tupleid,
 		   EState *estate)
 {
@@ -1252,12 +1245,13 @@ ExecAppend(TupleTableSlot *slot,
 	 * Check the constraints of the tuple
 	 */
 	if (resultRelationDesc->rd_att->constr)
-		ExecConstraints("ExecAppend", resultRelInfo, slot, estate);
+		ExecConstraints("ExecInsert", resultRelInfo, slot, estate);
 
 	/*
 	 * insert the tuple
 	 */
-	newId = heap_insert(resultRelationDesc, tuple);
+	newId = heap_insert(resultRelationDesc, tuple,
+						estate->es_snapshot->curcid);
 
 	IncrAppended();
 	(estate->es_processed)++;
@@ -1283,7 +1277,7 @@ ExecAppend(TupleTableSlot *slot,
 /* ----------------------------------------------------------------
  *		ExecDelete
  *
- *		DELETE is like append, we delete the tuple and its
+ *		DELETE is like UPDATE, we delete the tuple and its
  *		index tuples.
  * ----------------------------------------------------------------
  */
@@ -1294,8 +1288,9 @@ ExecDelete(TupleTableSlot *slot,
 {
 	ResultRelInfo *resultRelInfo;
 	Relation	resultRelationDesc;
-	ItemPointerData ctid;
 	int			result;
+	ItemPointerData update_ctid;
+	TransactionId update_xmax;
 
 	/*
 	 * get information on the (current) result relation
@@ -1309,7 +1304,8 @@ ExecDelete(TupleTableSlot *slot,
 	{
 		bool		dodelete;
 
-		dodelete = ExecBRDeleteTriggers(estate, resultRelInfo, tupleid);
+		dodelete = ExecBRDeleteTriggers(estate, resultRelInfo, tupleid,
+										estate->es_snapshot->curcid);
 
 		if (!dodelete)			/* "do nothing" */
 			return;
@@ -1319,10 +1315,13 @@ ExecDelete(TupleTableSlot *slot,
 	 * delete the tuple
 	 */
 ldelete:;
-	result = heap_delete(resultRelationDesc, tupleid, &ctid);
+	result = heap_delete(resultRelationDesc, tupleid,
+						 &update_ctid, &update_xmax,
+						 estate->es_snapshot->curcid);
 	switch (result)
 	{
 		case HeapTupleSelfUpdated:
+			/* already deleted by self; nothing to do */
 			return;
 
 		case HeapTupleMayBeUpdated:
@@ -1331,14 +1330,18 @@ ldelete:;
 		case HeapTupleUpdated:
 			if (XactIsoLevel == XACT_SERIALIZABLE)
 				elog(ERROR, "Can't serialize access due to concurrent update");
-			else if (!(ItemPointerEquals(tupleid, &ctid)))
+			else if (!ItemPointerEquals(tupleid, &update_ctid))
 			{
-				TupleTableSlot *epqslot = EvalPlanQual(estate,
-							   resultRelInfo->ri_RangeTableIndex, &ctid);
+				TupleTableSlot *epqslot;
 
+				epqslot = EvalPlanQual(estate,
+									   resultRelInfo->ri_RangeTableIndex,
+									   &update_ctid,
+									   update_xmax,
+									   estate->es_snapshot->curcid);
 				if (!TupIsNull(epqslot))
 				{
-					*tupleid = ctid;
+					*tupleid = update_ctid;
 					goto ldelete;
 				}
 			}
@@ -1368,26 +1371,27 @@ ldelete:;
 }
 
 /* ----------------------------------------------------------------
- *		ExecReplace
+ *		ExecUpdate
  *
- *		note: we can't run replace queries with transactions
- *		off because replaces are actually appends and our
- *		scan will mistakenly loop forever, replacing the tuple
- *		it just appended..	This should be fixed but until it
+ *		note: we can't run UPDATE queries with transactions
+ *		off because UPDATEs are actually INSERTs and our
+ *		scan will mistakenly loop forever, updating the tuple
+ *		it just inserted..	This should be fixed but until it
  *		is, we don't want to get stuck in an infinite loop
  *		which corrupts your database..
  * ----------------------------------------------------------------
  */
 static void
-ExecReplace(TupleTableSlot *slot,
-			ItemPointer tupleid,
-			EState *estate)
+ExecUpdate(TupleTableSlot *slot,
+		   ItemPointer tupleid,
+		   EState *estate)
 {
 	HeapTuple	tuple;
 	ResultRelInfo *resultRelInfo;
 	Relation	resultRelationDesc;
-	ItemPointerData ctid;
 	int			result;
+	ItemPointerData update_ctid;
+	TransactionId update_xmax;
 	int			numIndices;
 
 	/*
@@ -1395,7 +1399,7 @@ ExecReplace(TupleTableSlot *slot,
 	 */
 	if (IsBootstrapProcessingMode())
 	{
-		elog(NOTICE, "ExecReplace: replace can't run without transactions");
+		elog(WARNING, "ExecUpdate: UPDATE can't run without transactions");
 		return;
 	}
 
@@ -1417,7 +1421,8 @@ ExecReplace(TupleTableSlot *slot,
 		HeapTuple	newtuple;
 
 		newtuple = ExecBRUpdateTriggers(estate, resultRelInfo,
-										tupleid, tuple);
+										tupleid, tuple,
+										estate->es_snapshot->curcid);
 
 		if (newtuple == NULL)	/* "do nothing" */
 			return;
@@ -1446,15 +1451,18 @@ ExecReplace(TupleTableSlot *slot,
 	 */
 lreplace:;
 	if (resultRelationDesc->rd_att->constr)
-		ExecConstraints("ExecReplace", resultRelInfo, slot, estate);
+		ExecConstraints("ExecUpdate", resultRelInfo, slot, estate);
 
 	/*
 	 * replace the heap tuple
 	 */
-	result = heap_update(resultRelationDesc, tupleid, tuple, &ctid);
+	result = heap_update(resultRelationDesc, tupleid, tuple,
+						 &update_ctid, &update_xmax,
+						 estate->es_snapshot->curcid);
 	switch (result)
 	{
 		case HeapTupleSelfUpdated:
+			/* already deleted by self; nothing to do */
 			return;
 
 		case HeapTupleMayBeUpdated:
@@ -1463,14 +1471,18 @@ lreplace:;
 		case HeapTupleUpdated:
 			if (XactIsoLevel == XACT_SERIALIZABLE)
 				elog(ERROR, "Can't serialize access due to concurrent update");
-			else if (!(ItemPointerEquals(tupleid, &ctid)))
+			else if (!(ItemPointerEquals(tupleid, &update_ctid)))
 			{
-				TupleTableSlot *epqslot = EvalPlanQual(estate,
-							   resultRelInfo->ri_RangeTableIndex, &ctid);
+				TupleTableSlot *epqslot;
 
+				epqslot = EvalPlanQual(estate,
+									   resultRelInfo->ri_RangeTableIndex,
+									   &update_ctid,
+									   update_xmax,
+									   estate->es_snapshot->curcid);
 				if (!TupIsNull(epqslot))
 				{
-					*tupleid = ctid;
+					*tupleid = update_ctid;
 					tuple = ExecRemoveJunk(estate->es_junkFilter, epqslot);
 					slot = ExecStoreTuple(tuple,
 									estate->es_junkFilter->jf_resultSlot,
@@ -1492,7 +1504,7 @@ lreplace:;
 	/*
 	 * Note: instead of having to update the old index tuples associated
 	 * with the heap tuple, all we do is form and insert new index tuples.
-	 * This is because replaces are actually deletes and inserts and index
+	 * This is because UPDATEs are actually DELETEs and INSERTs and index
 	 * tuple deletion is done automagically by the vacuum daemon. All we
 	 * do is insert new index tuples.  -cim 9/27/89
 	 */
@@ -1501,7 +1513,7 @@ lreplace:;
 	 * process indices
 	 *
 	 * heap_update updates a tuple in the base relation by invalidating it
-	 * and then appending a new tuple to the relation.	As a side effect,
+	 * and then inserting a new tuple to the relation.	As a side effect,
 	 * the tupleid of the new tuple is placed in the new tuple's t_ctid
 	 * field.  So we now insert index tuples using the new tupleid stored
 	 * there.
@@ -1509,7 +1521,7 @@ lreplace:;
 
 	numIndices = resultRelInfo->ri_NumIndices;
 	if (numIndices > 0)
-		ExecInsertIndexTuples(slot, &(tuple->t_self), estate, true);
+		ExecInsertIndexTuples(slot, &(tuple->t_self), estate, false);
 
 	/* AFTER ROW UPDATE Triggers */
 	if (resultRelInfo->ri_TrigDesc)
@@ -1574,7 +1586,7 @@ ExecRelCheck(ResultRelInfo *resultRelInfo,
 }
 
 void
-ExecConstraints(char *caller, ResultRelInfo *resultRelInfo,
+ExecConstraints(const char *caller, ResultRelInfo *resultRelInfo,
 				TupleTableSlot *slot, EState *estate)
 {
 	Relation	rel = resultRelInfo->ri_RelationDesc;
@@ -1602,8 +1614,8 @@ ExecConstraints(char *caller, ResultRelInfo *resultRelInfo,
 		char	   *failed;
 
 		if ((failed = ExecRelCheck(resultRelInfo, slot, estate)) != NULL)
-			elog(ERROR, "%s: rejected due to CHECK constraint %s",
-				 caller, failed);
+			elog(ERROR, "%s: rejected due to CHECK constraint \"%s\" on \"%s\"",
+				 caller, failed, RelationGetRelationName(rel));
 	}
 }
 
@@ -1612,9 +1624,22 @@ ExecConstraints(char *caller, ResultRelInfo *resultRelInfo,
  * under READ COMMITTED rules.
  *
  * See backend/executor/README for some info about how this works.
+ *
+ *	estate - executor state data
+ *	rti - rangetable index of table containing tuple
+ *	*tid - t_ctid from the outdated tuple (ie, next updated version)
+ *	priorXmax - t_xmax from the outdated tuple
+ *	curCid - command ID of current command of my transaction
+ *
+ * *tid is also an output parameter: it's modified to hold the TID of the
+ * latest version of the tuple (note this may be changed even on failure)
+ *
+ * Returns a slot containing the new candidate update/delete tuple, or
+ * NULL if we determine we shouldn't process the row.
  */
 TupleTableSlot *
-EvalPlanQual(EState *estate, Index rti, ItemPointer tid)
+EvalPlanQual(EState *estate, Index rti,
+			 ItemPointer tid, TransactionId priorXmax, CommandId curCid)
 {
 	evalPlanQual *epq;
 	EState	   *epqstate;
@@ -1659,11 +1684,24 @@ EvalPlanQual(EState *estate, Index rti, ItemPointer tid)
 	{
 		Buffer		buffer;
 
-		heap_fetch(relation, SnapshotDirty, &tuple, &buffer, NULL);
-		if (tuple.t_data != NULL)
+		if (heap_fetch(relation, SnapshotDirty, &tuple, &buffer, true, NULL))
 		{
-			TransactionId xwait = SnapshotDirty->xmax;
+			/*
+			 * If xmin isn't what we're expecting, the slot must have been
+			 * recycled and reused for an unrelated tuple.  This implies
+			 * that the latest version of the row was deleted, so we need
+			 * do nothing.  (Should be safe to examine xmin without getting
+			 * buffer's content lock, since xmin never changes in an existing
+			 * tuple.)
+			 */
+			if (!TransactionIdEquals(HeapTupleHeaderGetXmin(tuple.t_data),
+									 priorXmax))
+			{
+				ReleaseBuffer(buffer);
+				return NULL;
+			}
 
+			/* otherwise xmin should not be dirty... */
 			if (TransactionIdIsValid(SnapshotDirty->xmin))
 				elog(ERROR, "EvalPlanQual: t_xmin is uncommitted ?!");
 
@@ -1671,11 +1709,29 @@ EvalPlanQual(EState *estate, Index rti, ItemPointer tid)
 			 * If tuple is being updated by other transaction then we have
 			 * to wait for its commit/abort.
 			 */
-			if (TransactionIdIsValid(xwait))
+			if (TransactionIdIsValid(SnapshotDirty->xmax))
 			{
 				ReleaseBuffer(buffer);
-				XactLockTableWait(xwait);
-				continue;
+				XactLockTableWait(SnapshotDirty->xmax);
+				continue;		/* loop back to repeat heap_fetch */
+			}
+
+			/*
+			 * If tuple was inserted by our own transaction, we have to check
+			 * cmin against curCid: cmin >= curCid means our command cannot
+			 * see the tuple, so we should ignore it.  Without this we are
+			 * open to the "Halloween problem" of indefinitely re-updating
+			 * the same tuple.  (We need not check cmax because
+			 * HeapTupleSatisfiesDirty will consider a tuple deleted by
+			 * our transaction dead, regardless of cmax.)  We just checked
+			 * that priorXmax == xmin, so we can test that variable instead
+			 * of doing HeapTupleHeaderGetXmin again.
+			 */
+			if (TransactionIdIsCurrentTransactionId(priorXmax) &&
+				HeapTupleHeaderGetCmin(tuple.t_data) >= curCid)
+			{
+				ReleaseBuffer(buffer);
+				return NULL;
 			}
 
 			/*
@@ -1687,22 +1743,50 @@ EvalPlanQual(EState *estate, Index rti, ItemPointer tid)
 		}
 
 		/*
-		 * Oops! Invalid tuple. Have to check is it updated or deleted.
-		 * Note that it's possible to get invalid SnapshotDirty->tid if
-		 * tuple updated by this transaction. Have we to check this ?
+		 * If the referenced slot was actually empty, the latest version
+		 * of the row must have been deleted, so we need do nothing.
 		 */
-		if (ItemPointerIsValid(&(SnapshotDirty->tid)) &&
-			!(ItemPointerEquals(&(tuple.t_self), &(SnapshotDirty->tid))))
+		if (tuple.t_data == NULL)
 		{
-			/* updated, so look at the updated copy */
-			tuple.t_self = SnapshotDirty->tid;
-			continue;
+			ReleaseBuffer(buffer);
+			return NULL;
 		}
 
 		/*
-		 * Deleted or updated by this transaction; forget it.
+		 * As above, if xmin isn't what we're expecting, do nothing.
 		 */
-		return NULL;
+		if (!TransactionIdEquals(HeapTupleHeaderGetXmin(tuple.t_data),
+								 priorXmax))
+		{
+			ReleaseBuffer(buffer);
+			return NULL;
+		}
+
+		/*
+		 * If we get here, the tuple was found but failed SnapshotDirty.
+		 * Assuming the xmin is either a committed xact or our own xact
+		 * (as it certainly should be if we're trying to modify the tuple),
+		 * this must mean that the row was updated or deleted by either
+		 * a committed xact or our own xact.  If it was deleted, we can
+		 * ignore it; if it was updated then chain up to the next version
+		 * and repeat the whole test.
+		 *
+		 * As above, it should be safe to examine xmax and t_ctid without
+		 * the buffer content lock, because they can't be changing.
+		 */
+		if (ItemPointerEquals(&tuple.t_self, &tuple.t_data->t_ctid))
+		{
+			/* deleted, so forget about it */
+			ReleaseBuffer(buffer);
+			return NULL;
+		}
+
+		/* updated, so look at the updated row */
+		tuple.t_self = tuple.t_data->t_ctid;
+		/* updated row should have xmin matching this xmax */
+		priorXmax = HeapTupleHeaderGetXmax(tuple.t_data);
+		ReleaseBuffer(buffer);
+		/* loop back to fetch next in chain */
 	}
 
 	/*
@@ -1865,6 +1949,7 @@ EvalPlanQual(EState *estate, Index rti, ItemPointer tid)
 	if (estate->es_origPlan->nParamExec > 0)
 		memset(epqstate->es_param_exec_vals, 0,
 			   estate->es_origPlan->nParamExec * sizeof(ParamExecData));
+	epqstate->es_force_oids = estate->es_force_oids;
 	memset(epqstate->es_evTupleNull, false, rtsize * sizeof(bool));
 	epqstate->es_useEvalPlan = false;
 	Assert(epqstate->es_tupleTable == NULL);
