@@ -1,16 +1,16 @@
 /* -----------------------------------------------------------------------
  * formatting.c
  *
- * $Header: /cvsroot/pgsql/src/backend/utils/adt/formatting.c,v 1.49.2.1 2005/03/26 00:42:56 tgl Exp $
+ * $Header: /cvsroot/pgsql/src/backend/utils/adt/formatting.c,v 1.56.2.4 2007/06/29 01:52:21 tgl Exp $
  *
  *
- *	 Portions Copyright (c) 1999-2000, PostgreSQL Global Development Group
+ *	 Portions Copyright (c) 1999-2002, PostgreSQL Global Development Group
  *
  *
  *	 TO_CHAR(); TO_TIMESTAMP(); TO_DATE(); TO_NUMBER();
  *
  *	 The PostgreSQL routines for a timestamp/int/float/numeric formatting,
- *	 inspire with Oracle TO_CHAR() / TO_DATE() / TO_NUMBER() routines.
+ *	 inspired by the Oracle TO_CHAR() / TO_DATE() / TO_NUMBER() routines.
  *
  *
  *	 Cache & Memory:
@@ -64,7 +64,7 @@
  */
 /***
 #define DEBUG_TO_FROM_CHAR
-#define DEBUG_elog_output	NOTICE
+#define DEBUG_elog_output	DEBUG3
 ***/
 
 #include "postgres.h"
@@ -72,11 +72,10 @@
 #include <ctype.h>
 #include <sys/time.h>
 #include <unistd.h>
-#ifdef USE_LOCALE
 #include <locale.h>
-#endif
 #include <math.h>
 #include <float.h>
+#include <limits.h>
 
 #include "utils/builtins.h"
 #include "utils/date.h"
@@ -111,8 +110,8 @@
  * More is in float.c
  * ----------
  */
-#define MAXFLOATWIDTH	64
-#define MAXDOUBLEWIDTH	128
+#define MAXFLOATWIDTH	60
+#define MAXDOUBLEWIDTH	500
 
 /* ----------
  * External (defined in PgSQL dt.c (timestamp utils))
@@ -412,7 +411,7 @@ typedef struct
 typedef struct TmToChar
 {
 	struct tm	tm;				/* classic 'tm' struct */
-	double		fsec;			/* milliseconds */
+	fsec_t		fsec;			/* fractional seconds */
 	char	   *tzn;			/* timezone */
 } TmToChar;
 
@@ -434,17 +433,6 @@ typedef struct TmToChar
 		tmtcTzn(_X) = NULL; \
 	} while(0)
 
-/* ----------
- * Utils
- * ----------
- */
-#ifndef MIN
-#define MIN(a,b) (((a)<(b)) ? (a) : (b))
-#endif
-#ifndef MAX
-#define MAX(a,b) (((a)>(b)) ? (a) : (b))
-#endif
-
 /*****************************************************************************
  *			KeyWords definition & action
  *****************************************************************************/
@@ -894,7 +882,7 @@ static int	dch_global(int arg, char *inout, int suf, int flag, FormatNode *node,
 static int	dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data);
 static int	dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data);
 static char *fill_str(char *str, int c, int max);
-static FormatNode *NUM_cache(int len, NUMDesc *Num, char *pars_str, int *flag);
+static FormatNode *NUM_cache(int len, NUMDesc *Num, char *pars_str, bool *shouldFree);
 static char *int_to_roman(int number);
 static void NUM_prepare_locale(NUMProc *Np);
 static char *get_last_relevant_decnum(char *num);
@@ -912,7 +900,7 @@ static void NUM_cache_remove(NUMCacheEntry *ent);
 
 /* ----------
  * Fast sequential search, use index for data selection which
- * go to seq. cycle (it is very fast for non-wanted strings)
+ * go to seq. cycle (it is very fast for unwanted strings)
  * (can't be used binary search in format parsing)
  * ----------
  */
@@ -970,7 +958,6 @@ NUMDesc_prepare(NUMDesc *num, FormatNode *n)
 
 	switch (n->key->id)
 	{
-
 		case NUM_9:
 			if (IS_BRACKET(num))
 			{
@@ -1278,6 +1265,16 @@ DCH_processor(FormatNode *node, char *inout, int flag, void *data)
 
 	for (n = node, s = inout; n->type != NODE_TYPE_END; n++)
 	{
+		if (flag == FROM_CHAR && *s=='\0')
+			/*
+			 * The input string is shorter than format picture, 
+			 * so it's good time to break this loop...
+			 * 
+			 * Note: this isn't relevant for TO_CHAR mode, beacuse 
+			 *       it use 'inout' allocated by format picture length.
+			 */
+			break;
+
 		if (n->type == NODE_TYPE_ACTION)
 		{
 			int			len;
@@ -1421,7 +1418,7 @@ str_numth(char *dest, char *num, int type)
 }
 
 /* ----------
- * Convert string to upper-string
+ * Convert string to upper case. Input string is modified in place.
  * ----------
  */
 static char *
@@ -1429,6 +1426,9 @@ str_toupper(char *buff)
 {
 	char	   *p_buff = buff;
 
+	if (!buff)
+		return NULL;
+
 	while (*p_buff)
 	{
 		*p_buff = toupper((unsigned char) *p_buff);
@@ -1438,7 +1438,7 @@ str_toupper(char *buff)
 }
 
 /* ----------
- * Convert string to lower-string
+ * Convert string to lower case. Input string is modified in place.
  * ----------
  */
 static char *
@@ -1446,6 +1446,9 @@ str_tolower(char *buff)
 {
 	char	   *p_buff = buff;
 
+	if (!buff)
+		return NULL;
+
 	while (*p_buff)
 	{
 		*p_buff = tolower((unsigned char) *p_buff);
@@ -1844,7 +1847,11 @@ dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_MS:			/* millisecond */
 			if (flag == TO_CHAR)
 			{
+#ifdef HAVE_INT64_TIMESTAMP
+				sprintf(inout, "%03d", (int) (tmtc->fsec / INT64CONST(1000)));
+#else
 				sprintf(inout, "%03d", (int) rint(tmtc->fsec * 1000));
+#endif
 				if (S_THth(suf))
 					str_numth(p_inout, inout, S_TH_TYPE(suf));
 				if (S_THth(suf))
@@ -1878,7 +1885,7 @@ dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 					x == 2 ? 10 : 1;
 
 				/*
-				 * elog(NOTICE, "X: %d, MS: %d, LEN: %d", x, tmfc->ms,
+				 * elog(DEBUG3, "X: %d, MS: %d, LEN: %d", x, tmfc->ms,
 				 * len);
 				 */
 				return len - 1 + SKIP_THth(suf);
@@ -1887,7 +1894,11 @@ dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_US:			/* microsecond */
 			if (flag == TO_CHAR)
 			{
+#ifdef HAVE_INT64_TIMESTAMP
+				sprintf(inout, "%06d", (int) tmtc->fsec);
+#else
 				sprintf(inout, "%06d", (int) rint(tmtc->fsec * 1000000));
+#endif
 				if (S_THth(suf))
 					str_numth(p_inout, inout, S_TH_TYPE(suf));
 				if (S_THth(suf))
@@ -1919,7 +1930,7 @@ dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 					x == 5 ? 10 : 1;
 
 				/*
-				 * elog(NOTICE, "X: %d, US: %d, LEN: %d", x, tmfc->us,
+				 * elog(DEBUG3, "X: %d, US: %d, LEN: %d", x, tmfc->us,
 				 * len);
 				 */
 				return len - 1 + SKIP_THth(suf);
@@ -1953,19 +1964,16 @@ dch_time(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_TZ:
 			if (flag == TO_CHAR && tmtcTzn(tmtc))
 			{
-				int			siz = strlen(tmtcTzn(tmtc));
-
 				if (arg == DCH_TZ)
 					strcpy(inout, tmtcTzn(tmtc));
 				else
 				{
-					char	   *p = palloc(siz);
+					char	   *p = pstrdup(tmtcTzn(tmtc));
 
-					strcpy(p, tmtcTzn(tmtc));
 					strcpy(inout, str_tolower(p));
 					pfree(p);
 				}
-				return siz - 1;
+				return strlen(inout) - 1;
 			}
 			else if (flag == FROM_CHAR)
 				elog(ERROR, "to_timestamp(): TZ/tz not supported.");
@@ -2439,7 +2447,7 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_YYY:
 			if (flag == TO_CHAR)
 			{
-				sprintf(buff, "%03d", YEAR_ABS(tm->tm_year));
+				snprintf(buff, sizeof(buff), "%03d", YEAR_ABS(tm->tm_year));
 				i = strlen(buff);
 				strcpy(inout, buff + (i - 3));
 				if (S_THth(suf))
@@ -2453,11 +2461,10 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 			else if (flag == FROM_CHAR)
 			{
 				sscanf(inout, "%03d", &tmfc->year);
-				
+
 				/*
-				 * 3-digit year:
-				 *	'100' ... '999' = 1100 ... 1999
-				 *	'000' ... '099' = 2000 ... 2099
+				 * 3-digit year: '100' ... '999' = 1100 ... 1999 '000' ...
+				 * '099' = 2000 ... 2099
 				 */
 				if (tmfc->year >= 100)
 					tmfc->year += 1000;
@@ -2470,7 +2477,7 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_YY:
 			if (flag == TO_CHAR)
 			{
-				sprintf(buff, "%02d", YEAR_ABS(tm->tm_year));
+				snprintf(buff, sizeof(buff), "%02d", YEAR_ABS(tm->tm_year));
 				i = strlen(buff);
 				strcpy(inout, buff + (i - 2));
 				if (S_THth(suf))
@@ -2486,10 +2493,9 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 				sscanf(inout, "%02d", &tmfc->year);
 
 				/*
-				 * 2-digit year:
-				 * '00' ... '69'  = 2000 ... 2069
-				 * '70' ... '99'  = 1970 ... 1999
-		 		 */
+				 * 2-digit year: '00' ... '69'	= 2000 ... 2069 '70' ...
+				 * '99'  = 1970 ... 1999
+				 */
 				if (tmfc->year < 70)
 					tmfc->year += 2000;
 				else
@@ -2501,7 +2507,7 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 		case DCH_Y:
 			if (flag == TO_CHAR)
 			{
-				sprintf(buff, "%1d", YEAR_ABS(tm->tm_year));
+				snprintf(buff, sizeof(buff), "%1d", YEAR_ABS(tm->tm_year));
 				i = strlen(buff);
 				strcpy(inout, buff + (i - 1));
 				if (S_THth(suf))
@@ -2515,7 +2521,7 @@ dch_date(int arg, char *inout, int suf, int flag, FormatNode *node, void *data)
 			else if (flag == FROM_CHAR)
 			{
 				sscanf(inout, "%1d", &tmfc->year);
-				
+
 				/*
 				 * 1-digit year: always +2000
 				 */
@@ -2610,7 +2616,7 @@ DCH_cache_getnew(char *str)
 {
 	DCHCacheEntry *ent = NULL;
 
-	/* counter overload check  - paranoa? */
+	/* counter overload check  - paranoia? */
 	if (DCHCounter + DCH_CACHE_FIELDS >= MAX_INT32)
 	{
 		DCHCounter = 0;
@@ -2638,7 +2644,7 @@ DCH_cache_getnew(char *str)
 #ifdef DEBUG_TO_FROM_CHAR
 		elog(DEBUG_elog_output, "OLD: '%s' AGE: %d", old->str, old->age);
 #endif
-		strcpy(old->str, str);	/* check str size before this func. */
+		StrNCpy(old->str, str, DCH_CACHE_SIZE + 1);
 		/* old->format fill parser */
 		old->age = (++DCHCounter);
 		return old;
@@ -2650,7 +2656,7 @@ DCH_cache_getnew(char *str)
 		elog(DEBUG_elog_output, "NEW (%d)", n_DCHCache);
 #endif
 		ent = DCHCache + n_DCHCache;
-		strcpy(ent->str, str);	/* check str size before this func. */
+		StrNCpy(ent->str, str, DCH_CACHE_SIZE + 1);
 		/* ent->format fill parser */
 		ent->age = (++DCHCounter);
 		++n_DCHCache;
@@ -2666,7 +2672,7 @@ DCH_cache_search(char *str)
 	int			i = 0;
 	DCHCacheEntry *ent;
 
-	/* counter overload check  - paranoa? */
+	/* counter overload check  - paranoia? */
 	if (DCHCounter + DCH_CACHE_FIELDS >= MAX_INT32)
 	{
 		DCHCounter = 0;
@@ -2729,7 +2735,7 @@ datetime_to_char_body(TmToChar *tmtc, text *fmt)
 		parse_format(format, str_fmt, DCH_keywords,
 					 DCH_suff, DCH_index, DCH_TYPE, NULL);
 
-		(format + len)->type = NODE_TYPE_END;	/* Paranoa? */
+		(format + len)->type = NODE_TYPE_END;	/* Paranoia? */
 
 	}
 	else
@@ -2753,7 +2759,7 @@ datetime_to_char_body(TmToChar *tmtc, text *fmt)
 			parse_format(ent->format, str_fmt, DCH_keywords,
 						 DCH_suff, DCH_index, DCH_TYPE, NULL);
 
-			(ent->format + len)->type = NODE_TYPE_END;	/* Paranoa? */
+			(ent->format + len)->type = NODE_TYPE_END;	/* Paranoia? */
 
 #ifdef DEBUG_TO_FROM_CHAR
 			/* dump_node(ent->format, len); */
@@ -2897,7 +2903,7 @@ to_timestamp(PG_FUNCTION_ARGS)
 				date_len,
 				tz = 0;
 	struct tm	tm;
-	double		fsec = 0;
+	fsec_t		fsec = 0;
 
 	ZERO_tm(&tm);
 	ZERO_tmfc(&tmfc);
@@ -2923,7 +2929,7 @@ to_timestamp(PG_FUNCTION_ARGS)
 			parse_format(format, str, DCH_keywords,
 						 DCH_suff, DCH_index, DCH_TYPE, NULL);
 
-			(format + len)->type = NODE_TYPE_END;		/* Paranoa? */
+			(format + len)->type = NODE_TYPE_END;		/* Paranoia? */
 		}
 		else
 		{
@@ -2932,7 +2938,7 @@ to_timestamp(PG_FUNCTION_ARGS)
 			 */
 			DCHCacheEntry *ent;
 
-			incache = 0;
+			incache = TRUE;
 
 			if ((ent = DCH_cache_search(str)) == NULL)
 			{
@@ -2946,7 +2952,7 @@ to_timestamp(PG_FUNCTION_ARGS)
 				parse_format(ent->format, str, DCH_keywords,
 							 DCH_suff, DCH_index, DCH_TYPE, NULL);
 
-				(ent->format + len)->type = NODE_TYPE_END;		/* Paranoa? */
+				(ent->format + len)->type = NODE_TYPE_END;		/* Paranoia? */
 #ifdef DEBUG_TO_FROM_CHAR
 				/* dump_node(ent->format, len); */
 				/* dump_index(DCH_keywords, DCH_index); */
@@ -2974,7 +2980,7 @@ to_timestamp(PG_FUNCTION_ARGS)
 
 		pfree(date_str);
 		pfree(str);
-		if (incache)
+		if (!incache)
 			pfree(format);
 	}
 
@@ -3100,10 +3106,17 @@ to_timestamp(PG_FUNCTION_ARGS)
 				tm.tm_yday - y[i - 1];
 	}
 
+#ifdef HAVE_INT64_TIMESTAMP
+	if (tmfc.ms)
+		fsec += tmfc.ms * 1000;
+	if (tmfc.us)
+		fsec += tmfc.us;
+#else
 	if (tmfc.ms)
 		fsec += (double) tmfc.ms / 1000;
 	if (tmfc.us)
 		fsec += (double) tmfc.us / 1000000;
+#endif
 
 	/* -------------------------------------------------------------- */
 
@@ -3140,7 +3153,7 @@ static char *
 fill_str(char *str, int c, int max)
 {
 	memset(str, c, max);
-	*(str + max + 1) = '\0';
+	*(str + max) = '\0';
 	return str;
 }
 
@@ -3162,7 +3175,7 @@ NUM_cache_getnew(char *str)
 {
 	NUMCacheEntry *ent = NULL;
 
-	/* counter overload check  - paranoa? */
+	/* counter overload check  - paranoia? */
 	if (NUMCounter + NUM_CACHE_FIELDS >= MAX_INT32)
 	{
 		NUMCounter = 0;
@@ -3199,7 +3212,7 @@ NUM_cache_getnew(char *str)
 #ifdef DEBUG_TO_FROM_CHAR
 		elog(DEBUG_elog_output, "OLD: '%s' AGE: %d", old->str, old->age);
 #endif
-		strcpy(old->str, str);	/* check str size before this func. */
+		StrNCpy(old->str, str, NUM_CACHE_SIZE + 1);
 		/* old->format fill parser */
 		old->age = (++NUMCounter);
 
@@ -3212,7 +3225,7 @@ NUM_cache_getnew(char *str)
 		elog(DEBUG_elog_output, "NEW (%d)", n_NUMCache);
 #endif
 		ent = NUMCache + n_NUMCache;
-		strcpy(ent->str, str);	/* check str size before this func. */
+		StrNCpy(ent->str, str, NUM_CACHE_SIZE + 1);
 		/* ent->format fill parser */
 		ent->age = (++NUMCounter);
 		++n_NUMCache;
@@ -3230,7 +3243,7 @@ NUM_cache_search(char *str)
 	int			i = 0;
 	NUMCacheEntry *ent;
 
-	/* counter overload check  - paranoa? */
+	/* counter overload check - paranoia? */
 	if (NUMCounter + NUM_CACHE_FIELDS >= MAX_INT32)
 	{
 		NUMCounter = 0;
@@ -3270,7 +3283,7 @@ NUM_cache_remove(NUMCacheEntry *ent)
  * ----------
  */
 static FormatNode *
-NUM_cache(int len, NUMDesc *Num, char *pars_str, int *flag)
+NUM_cache(int len, NUMDesc *Num, char *pars_str, bool *shouldFree)
 {
 	FormatNode *format = NULL;
 	char	   *str;
@@ -3284,20 +3297,21 @@ NUM_cache(int len, NUMDesc *Num, char *pars_str, int *flag)
 
 	/*
 	 * Allocate new memory if format picture is bigger than static cache
-	 * and not use cache (call parser always) - flag=1 show this variant
+	 * and not use cache (call parser always). This branches sets
+	 * shouldFree to true, accordingly.
 	 */
 	if (len > NUM_CACHE_SIZE)
 	{
-
 		format = (FormatNode *) palloc((len + 1) * sizeof(FormatNode));
-		*flag = 1;
+
+		*shouldFree = true;
 
 		zeroize_NUM(Num);
 
 		parse_format(format, str, NUM_keywords,
 					 NULL, NUM_index, NUM_TYPE, Num);
 
-		(format + len)->type = NODE_TYPE_END;	/* Paranoa? */
+		(format + len)->type = NODE_TYPE_END;	/* Paranoia? */
 
 	}
 	else
@@ -3307,7 +3321,7 @@ NUM_cache(int len, NUMDesc *Num, char *pars_str, int *flag)
 		 */
 		NUMCacheEntry *ent;
 
-		flag = 0;
+		*shouldFree = false;
 
 		if ((ent = NUM_cache_search(str)) == NULL)
 		{
@@ -3321,7 +3335,7 @@ NUM_cache(int len, NUMDesc *Num, char *pars_str, int *flag)
 			parse_format(ent->format, str, NUM_keywords,
 						 NULL, NUM_index, NUM_TYPE, &ent->Num);
 
-			(ent->format + len)->type = NODE_TYPE_END;	/* Paranoa? */
+			(ent->format + len)->type = NODE_TYPE_END;	/* Paranoia? */
 
 		}
 
@@ -3407,9 +3421,6 @@ int_to_roman(int number)
 static void
 NUM_prepare_locale(NUMProc *Np)
 {
-
-#ifdef USE_LOCALE
-
 	if (Np->Num->need_locale)
 	{
 
@@ -3463,8 +3474,6 @@ NUM_prepare_locale(NUMProc *Np)
 	}
 	else
 	{
-#endif
-
 		/*
 		 * Default values
 		 */
@@ -3473,10 +3482,7 @@ NUM_prepare_locale(NUMProc *Np)
 		Np->decimal = ".";
 		Np->L_thousands_sep = ",";
 		Np->L_currency_symbol = " ";
-
-#ifdef USE_LOCALE
 	}
-#endif
 }
 
 /* ----------
@@ -3785,11 +3791,11 @@ NUM_numpart_to_char(NUMProc *Np, int id)
 					strcpy(Np->inout_p, Np->decimal);	/* Write DEC/D */
 					Np->inout_p += strlen(Np->inout_p);
 
-					/*
-					 * terrible Ora '0' -- 9.9 --> '0.'
-					 */
 				}
-				else if (IS_FILLMODE(Np->Num) && *Np->number == '0' &&
+				/*
+				 * Ora 'n' -- FM9.9 --> 'n.'
+				 */
+				else if (IS_FILLMODE(Np->Num)  &&
 						 Np->last_relevant && *Np->last_relevant == '.')
 				{
 
@@ -3806,34 +3812,31 @@ NUM_numpart_to_char(NUMProc *Np, int id)
 				if (Np->last_relevant && Np->number_p > Np->last_relevant &&
 					id != NUM_0)
 					;
-
 				/*
 				 * terrible Ora format: '0.1' -- 9.9 --> '  .1'
 				 */
 				else if (!IS_ZERO(Np->Num) && *Np->number == '0' &&
 						 Np->number == Np->number_p && Np->Num->post != 0)
 				{
-
 					if (!IS_FILLMODE(Np->Num))
 					{
 						*Np->inout_p = ' ';
 						++Np->inout_p;
-
-						/*
-						 * total terible Ora: '0' -- FM9.9 --> '0.'
-						 */
 					}
+					/*
+					 * total terrible Ora: '0' -- FM9.9 --> '0.'
+					 */
 					else if (Np->last_relevant && *Np->last_relevant == '.')
 					{
 						*Np->inout_p = '0';
 						++Np->inout_p;
 					}
-
 				}
 				else
 				{
 #ifdef DEBUG_TO_FROM_CHAR
-					elog(DEBUG_elog_output, "Writing digit '%c' to position %d", *Np->number_p, Np->num_curr);
+					elog(DEBUG_elog_output, "Writing digit '%c' to position %d",
+						 *Np->number_p, Np->num_curr);
 #endif
 					*Np->inout_p = *Np->number_p;		/* Write DIGIT */
 					++Np->inout_p;
@@ -4281,10 +4284,10 @@ NUM_processor(FormatNode *node, NUMDesc *Num, char *inout, char *number,
 #define NUM_TOCHAR_prepare \
 do { \
 	len = VARSIZE(fmt) - VARHDRSZ;					\
-	if (len <= 0)							\
+	if (len <= 0 || len >= (INT_MAX-VARHDRSZ)/NUM_MAX_ITEM_SIZ)		\
 		return DirectFunctionCall1(textin, CStringGetDatum(""));	\
-	result	= (text *) palloc( (len * NUM_MAX_ITEM_SIZ) + 1 + VARHDRSZ); \
-	format	= NUM_cache(len, &Num, VARDATA(fmt), &flag);		\
+	result	= (text *) palloc((len * NUM_MAX_ITEM_SIZ) + 1 + VARHDRSZ);	\
+	format	= NUM_cache(len, &Num, VARDATA(fmt), &shouldFree);		\
 } while (0)
 
 /* ----------
@@ -4295,28 +4298,18 @@ do { \
 do { \
 	NUM_processor(format, &Num, VARDATA(result),			\
 		numstr, plen, sign, TO_CHAR);				\
-	pfree(orgnum);							\
 									\
-	if (flag)							\
-		pfree(format);						\
+	if (shouldFree)					\
+		pfree(format);				\
 									\
-	/*
-	 * for result is allocated max memory, which current format-picture\
-	 * needs, now it must be re-allocate to result real size	\
+	/*								\
+	 * Convert null-terminated representation of result to standard text. \
+	 * The result is usually much bigger than it needs to be, but there \
+	 * seems little point in realloc'ing it smaller. \
 	 */								\
-	if (!(len = strlen(VARDATA(result))))				\
-	{								\
-		pfree(result);						\
-		PG_RETURN_NULL();					\
-	}								\
-									\
-	result_tmp	= result;					\
-	result		= (text *) palloc( len + 1 + VARHDRSZ);		\
-									\
-	strcpy( VARDATA(result), VARDATA(result_tmp));			\
-	VARATT_SIZEP(result) = len + VARHDRSZ;				\
-	pfree(result_tmp);						\
-} while(0)
+	len = strlen(VARDATA(result));	\
+	VARATT_SIZEP(result) = len + VARHDRSZ;	\
+} while (0)
 
 /* -------------------
  * NUMERIC to_number() (convert string to numeric)
@@ -4331,17 +4324,17 @@ numeric_to_number(PG_FUNCTION_ARGS)
 	Datum		result;
 	FormatNode *format;
 	char	   *numstr;
-	int			flag = 0;
+	bool		shouldFree;
 	int			len = 0;
 	int			scale,
 				precision;
 
 	len = VARSIZE(fmt) - VARHDRSZ;
 
-	if (len <= 0)
+	if (len <= 0 || len >= INT_MAX/NUM_MAX_ITEM_SIZ)
 		PG_RETURN_NULL();
 
-	format = NUM_cache(len, &Num, VARDATA(fmt), &flag);
+	format = NUM_cache(len, &Num, VARDATA(fmt), &shouldFree);
 
 	numstr = (char *) palloc((len * NUM_MAX_ITEM_SIZ) + 1);
 
@@ -4349,9 +4342,9 @@ numeric_to_number(PG_FUNCTION_ARGS)
 				  VARSIZE(value) - VARHDRSZ, 0, FROM_CHAR);
 
 	scale = Num.post;
-	precision = MAX(0, Num.pre) + scale;
+	precision = Max(0, Num.pre) + scale;
 
-	if (flag)
+	if (shouldFree)
 		pfree(format);
 
 	result = DirectFunctionCall3(numeric_in,
@@ -4373,9 +4366,8 @@ numeric_to_char(PG_FUNCTION_ARGS)
 	text	   *fmt = PG_GETARG_TEXT_P(1);
 	NUMDesc		Num;
 	FormatNode *format;
-	text	   *result,
-			   *result_tmp;
-	int			flag = 0;
+	text	   *result;
+	bool		shouldFree;
 	int			len = 0,
 				plen = 0,
 				sign = 0;
@@ -4397,7 +4389,6 @@ numeric_to_char(PG_FUNCTION_ARGS)
 		numstr = orgnum =
 			int_to_roman(DatumGetInt32(DirectFunctionCall1(numeric_int4,
 												   NumericGetDatum(x))));
-		pfree(x);
 	}
 	else
 	{
@@ -4416,9 +4407,6 @@ numeric_to_char(PG_FUNCTION_ARGS)
 			val = DatumGetNumeric(DirectFunctionCall2(numeric_mul,
 												  NumericGetDatum(value),
 													NumericGetDatum(x)));
-			pfree(x);
-			pfree(a);
-			pfree(b);
 			Num.pre += Num.multi;
 		}
 
@@ -4427,10 +4415,9 @@ numeric_to_char(PG_FUNCTION_ARGS)
 												Int32GetDatum(Num.post)));
 		orgnum = DatumGetCString(DirectFunctionCall1(numeric_out,
 													 NumericGetDatum(x)));
-		pfree(x);
 
 		if (*orgnum == '-')
-		{						/* < 0 */
+		{
 			sign = '-';
 			numstr = orgnum + 1;
 		}
@@ -4449,13 +4436,10 @@ numeric_to_char(PG_FUNCTION_ARGS)
 
 		else if (len > Num.pre)
 		{
-			fill_str(numstr, '#', Num.pre);
+			numstr = (char *) palloc(Num.pre + Num.post + 2);
+			fill_str(numstr, '#', Num.pre + Num.post + 1);
 			*(numstr + Num.pre) = '.';
-			fill_str(numstr + 1 + Num.pre, '#', Num.post);
 		}
-
-		if (IS_MULTI(&Num))
-			pfree(val);
 	}
 
 	NUM_TOCHAR_finish;
@@ -4473,9 +4457,8 @@ int4_to_char(PG_FUNCTION_ARGS)
 	text	   *fmt = PG_GETARG_TEXT_P(1);
 	NUMDesc		Num;
 	FormatNode *format;
-	text	   *result,
-			   *result_tmp;
-	int			flag = 0;
+	text	   *result;
+	bool		shouldFree;
 	int			len = 0,
 				plen = 0,
 				sign = 0;
@@ -4502,40 +4485,34 @@ int4_to_char(PG_FUNCTION_ARGS)
 			orgnum = DatumGetCString(DirectFunctionCall1(int4out,
 												  Int32GetDatum(value)));
 		}
-		len = strlen(orgnum);
 
 		if (*orgnum == '-')
-		{						/* < 0 */
+		{
 			sign = '-';
-			--len;
+			orgnum++;
 		}
 		else
 			sign = '+';
+		len = strlen(orgnum);
 
 		if (Num.post)
 		{
-			int			i;
-
 			numstr = (char *) palloc(len + Num.post + 2);
-			strcpy(numstr, orgnum + (*orgnum == '-' ? 1 : 0));
+			strcpy(numstr, orgnum);
 			*(numstr + len) = '.';
-
-			for (i = len + 1; i <= len + Num.post; i++)
-				*(numstr + i) = '0';
+			memset(numstr + len + 1, '0', Num.post);
 			*(numstr + len + Num.post + 1) = '\0';
-			pfree(orgnum);
-			orgnum = numstr;
 		}
 		else
-			numstr = orgnum + (*orgnum == '-' ? 1 : 0);
+			numstr = orgnum;
 
 		if (Num.pre > len)
 			plen = Num.pre - len;
 		else if (len > Num.pre)
 		{
-			fill_str(numstr, '#', Num.pre);
+			numstr = (char *) palloc(Num.pre + Num.post + 2);
+			fill_str(numstr, '#', Num.pre + Num.post + 1);
 			*(numstr + Num.pre) = '.';
-			fill_str(numstr + 1 + Num.pre, '#', Num.post);
 		}
 	}
 
@@ -4554,9 +4531,8 @@ int8_to_char(PG_FUNCTION_ARGS)
 	text	   *fmt = PG_GETARG_TEXT_P(1);
 	NUMDesc		Num;
 	FormatNode *format;
-	text	   *result,
-			   *result_tmp;
-	int			flag = 0;
+	text	   *result;
+	bool		shouldFree;
 	int			len = 0,
 				plen = 0,
 				sign = 0;
@@ -4589,40 +4565,34 @@ int8_to_char(PG_FUNCTION_ARGS)
 
 		orgnum = DatumGetCString(DirectFunctionCall1(int8out,
 												  Int64GetDatum(value)));
-		len = strlen(orgnum);
 
 		if (*orgnum == '-')
-		{						/* < 0 */
+		{
 			sign = '-';
-			--len;
+			orgnum++;
 		}
 		else
 			sign = '+';
+		len = strlen(orgnum);
 
 		if (Num.post)
 		{
-			int			i;
-
 			numstr = (char *) palloc(len + Num.post + 2);
-			strcpy(numstr, orgnum + (*orgnum == '-' ? 1 : 0));
+			strcpy(numstr, orgnum);
 			*(numstr + len) = '.';
-
-			for (i = len + 1; i <= len + Num.post; i++)
-				*(numstr + i) = '0';
+			memset(numstr + len + 1, '0', Num.post);
 			*(numstr + len + Num.post + 1) = '\0';
-			pfree(orgnum);
-			orgnum = numstr;
 		}
 		else
-			numstr = orgnum + (*orgnum == '-' ? 1 : 0);
+			numstr = orgnum;
 
 		if (Num.pre > len)
 			plen = Num.pre - len;
 		else if (len > Num.pre)
 		{
-			fill_str(numstr, '#', Num.pre);
+			numstr = (char *) palloc(Num.pre + Num.post + 2);
+			fill_str(numstr, '#', Num.pre + Num.post + 1);
 			*(numstr + Num.pre) = '.';
-			fill_str(numstr + 1 + Num.pre, '#', Num.post);
 		}
 	}
 
@@ -4641,9 +4611,8 @@ float4_to_char(PG_FUNCTION_ARGS)
 	text	   *fmt = PG_GETARG_TEXT_P(1);
 	NUMDesc		Num;
 	FormatNode *format;
-	text	   *result,
-			   *result_tmp;
-	int			flag = 0;
+	text	   *result;
+	bool		shouldFree;
 	int			len = 0,
 				plen = 0,
 				sign = 0;
@@ -4671,7 +4640,7 @@ float4_to_char(PG_FUNCTION_ARGS)
 		}
 
 		orgnum = (char *) palloc(MAXFLOATWIDTH + 1);
-		sprintf(orgnum, "%.0f", fabs(val));
+		snprintf(orgnum, MAXFLOATWIDTH + 1, "%.0f", fabs(val));
 		len = strlen(orgnum);
 		if (Num.pre > len)
 			plen = Num.pre - len;
@@ -4679,7 +4648,7 @@ float4_to_char(PG_FUNCTION_ARGS)
 			Num.post = 0;
 		else if (Num.post + len > FLT_DIG)
 			Num.post = FLT_DIG - len;
-		sprintf(orgnum, "%.*f", Num.post, val);
+		snprintf(orgnum, MAXFLOATWIDTH + 1, "%.*f", Num.post, val);
 
 		if (*orgnum == '-')
 		{						/* < 0 */
@@ -4701,9 +4670,9 @@ float4_to_char(PG_FUNCTION_ARGS)
 
 		else if (len > Num.pre)
 		{
-			fill_str(numstr, '#', Num.pre);
+			numstr = (char *) palloc(Num.pre + Num.post + 2);
+			fill_str(numstr, '#', Num.pre + Num.post + 1);
 			*(numstr + Num.pre) = '.';
-			fill_str(numstr + 1 + Num.pre, '#', Num.post);
 		}
 	}
 
@@ -4722,9 +4691,8 @@ float8_to_char(PG_FUNCTION_ARGS)
 	text	   *fmt = PG_GETARG_TEXT_P(1);
 	NUMDesc		Num;
 	FormatNode *format;
-	text	   *result,
-			   *result_tmp;
-	int			flag = 0;
+	text	   *result;
+	bool		shouldFree;
 	int			len = 0,
 				plen = 0,
 				sign = 0;
@@ -4751,14 +4719,14 @@ float8_to_char(PG_FUNCTION_ARGS)
 			Num.pre += Num.multi;
 		}
 		orgnum = (char *) palloc(MAXDOUBLEWIDTH + 1);
-		len = sprintf(orgnum, "%.0f", fabs(val));
+		len = snprintf(orgnum, MAXDOUBLEWIDTH + 1, "%.0f", fabs(val));
 		if (Num.pre > len)
 			plen = Num.pre - len;
 		if (len >= DBL_DIG)
 			Num.post = 0;
 		else if (Num.post + len > DBL_DIG)
 			Num.post = DBL_DIG - len;
-		sprintf(orgnum, "%.*f", Num.post, val);
+		snprintf(orgnum, MAXDOUBLEWIDTH + 1, "%.*f", Num.post, val);
 
 		if (*orgnum == '-')
 		{						/* < 0 */
@@ -4780,9 +4748,9 @@ float8_to_char(PG_FUNCTION_ARGS)
 
 		else if (len > Num.pre)
 		{
-			fill_str(numstr, '#', Num.pre);
+			numstr = (char *) palloc(Num.pre + Num.post + 2);
+			fill_str(numstr, '#', Num.pre + Num.post + 1);
 			*(numstr + Num.pre) = '.';
-			fill_str(numstr + 1 + Num.pre, '#', Num.post);
 		}
 	}
 
