@@ -3,7 +3,7 @@
  *			  procedural language
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/pl/plpgsql/src/pl_comp.c,v 1.53.2.1 2003/01/31 00:32:00 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/pl/plpgsql/src/pl_comp.c,v 1.53.2.2 2007/02/08 18:38:28 tgl Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -504,12 +504,7 @@ plpgsql_compile(Oid fn_oid, int functype)
 	function->found_varno = var->varno;
 
 	/*
-	 * Forget about the above created variables
-	 */
-	plpgsql_add_initdatums(NULL);
-
-	/*
-	 * Now parse the functions text
+	 * Now parse the function's text
 	 */
 	parse_rc = plpgsql_yyparse();
 	if (parse_rc != 0)
@@ -1477,11 +1472,17 @@ plpgsql_adddatum(PLpgSQL_datum * new)
 
 
 /* ----------
- * plpgsql_add_initdatums		Put all datum entries created
- *					since the last call into the
- *					finishing code block so the
- *					block knows which variables to
- *					reinitialize when entered.
+ * plpgsql_add_initdatums		Make an array of the datum numbers of
+ *					all the simple VAR datums created since the last call
+ *					to this function.
+ *
+ * If varnos is NULL, we just forget any datum entries created since the
+ * last call.
+ *
+ * This is used around a DECLARE section to create a list of the VARs
+ * that have to be initialized at block entry.  Note that VARs can also
+ * be created elsewhere than DECLARE, eg by a FOR-loop, but it is then
+ * the responsibility of special-purpose code to initialize them.
  * ----------
  */
 int
